"""Library for creating beautiful and insightful visualizations."""

from importlib.metadata import version

from plastik import colors, lines
from plastik.axes import *  # noqa:F401,F403
from plastik.grid import *  # noqa:F401,F403
from plastik.legends import *  # noqa:F401,F403
from plastik.percentiles import percentiles
from plastik.ridge import *  # noqa:F401,F403

try:
    from plastik.airport import Airport, airport
except ImportError:

    def need_extra(*_, **k):
        raise ImportError(
            "Please install with extra dependencies: pip install plastik[extra]"
        )

    Airport = need_extra  # type: ignore[misc,assignment]
    airport = need_extra  # type: ignore[assignment]

__version__ = version(__package__)
__all__ = ["Airport", "airport", "colors", "lines", "percentiles"]



"""Create and manipulate grids, similar to sub-figure layouts."""

import itertools
from typing import Any, Literal

import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from typing_extensions import Self


class FigureGrid:
    """Return a figure with axes appropriate for (rows, columns) sub-figures.

    Parameters
    ----------
    rows : int | None
        The number of rows in the figure
    columns : int | None
        The number of columns in the figure
    """

    def __init__(
        self: Self,
        rows: int | None = None,
        columns: int | None = None,
    ) -> None:
        self.rows = 1 if rows is None else rows
        self.columns = 1 if columns is None else columns
        self._labels: list[str] | None = None
        self._pos: tuple[float, float] = (-0.2, 0.95)
        self._share_axes: Literal["x", "y", "both"] | None = None
        self._columns_first: bool = False
        self._expand_top: float = 1.0

    def __call__(
        self: Self,
        rows: int,
        columns: int,
        using: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> tuple[Figure, list[Axes]]:
        """Create a figure with sub-figures in the given layout.

        Parameters
        ----------
        rows : int
            The number of rows in the figure.
        columns : int
            The number of columns in the figure.
        using : dict[str, Any] | None
            The properties to be used for the sub-figures. The keys are the keyword
            arguments used in the `using` method, see its signature for accepted values.
        **kwargs : Any
            Additional keyword arguments to be passed to `Axes.text`.

        Returns
        -------
        Figure
            A new matplotlib figure object.
        list[Axes]
            A list of all the axes objects owned by the figure.
        """
        self.rows, self.columns = rows, columns
        if using is not None:
            self.using(**using)
        return self.get_grid(**kwargs)

    def _calculate_figsize(self: Self) -> tuple[float, float]:
        """Calculate the figure size based on the number of rows and columns."""
        full_cols = 3.37 * self.columns
        squash_cols = 3.37 * self.columns - (self.columns - 1) * 3.37 * 0.25
        full_rows = 2.08277 * self.rows
        squash_rows = 2.08277 * self.rows - (self.rows - 1) * 2.08277 * 0.25
        match self._share_axes:
            case None:
                return full_rows, full_cols
            case "x":
                return squash_rows, full_cols
            case "y":
                return full_rows, squash_cols
            case "both":
                return squash_rows, squash_cols
            case _:
                raise ValueError(f"Unknown value for share_axes: {self._share_axes}")

    def _update_labels(self: Self) -> list[str]:
        labels = (
            [
                rf"$\mathrm{{({chr(97 + ell)})}}$"
                for ell in range(self.rows * self.columns)
            ]
            if not self._labels or len(self._labels) != int(self.rows * self.columns)
            else self._labels.copy()
        )
        return self._maybe_columns_first(labels)

    def _maybe_columns_first(
        self: Self, list_: list, *, transpose: bool = True
    ) -> list:
        if not self._columns_first:
            return list_
        if transpose:
            list_ = [
                list_[i * self.rows + j]
                for j in range(self.rows)
                for i in range(self.columns)
            ]
            return list_
        grid = [
            list_[i * self.columns : (i + 1) * self.columns] for i in range(self.rows)
        ]
        return [
            grid[row][col]
            for col, row in itertools.product(range(self.columns), range(self.rows))
        ]

    def using(  # noqa: PLR0913
        self: Self,
        *,
        labels: list[str] | None = None,
        pos: tuple[float, float] | None = None,
        share_axes: Literal["x", "y", "both"] | None = None,
        columns_first: bool | None = None,
        expand_top: float | None = None,
    ) -> Self:
        """Set text properties.

        The properties must be given as keyword arguments to take effect.

        Parameters
        ----------
        labels : list[str] | None
            The labels to be applied to each sub-figure. Defaults to (a), (b), (c), ...
        pos : tuple[float, float] | None
            The position in the sub-figure relative to the bottom-left corner. Default
            is `(-0.2, 0.95)`.
        share_axes : Literal["x", "y", "both"] | None
            Use a shared axis for the given direction. Default is `None`.
        columns_first : bool | None
            If the labels should be placed in a columns-first order. Default is `False`,
            meaning rows are numbered first.
        expand_top : float | None
            Make the figure higher by multiplying by `expand_top`. Note that the
            subfigures will remain the same. Useful for placing a common legend at the
            top of the figure.

        Returns
        -------
        Self
            The object itself.
        """
        self._labels = labels or self._labels
        self._pos = pos or self._pos
        self._share_axes = share_axes or self._share_axes
        self._columns_first = columns_first or self._columns_first
        self._expand_top = expand_top or self._expand_top
        return self

    def get_grid(
        self: Self,
        **kwargs: dict,
    ) -> tuple[Figure, list[Axes]]:
        """Return a figure with axes appropriate for (rows, columns) sub-figures.

        Parameters
        ----------
        **kwargs : dict
            Additional keyword arguments to be passed to Axes.text.

        Returns
        -------
        Figure
            The figure object
        list[Axes]
            A list of all the axes objects owned by the figure
        """
        full_height, full_width = self._calculate_figsize()
        fig = plt.figure(figsize=(full_width, full_height * self._expand_top))
        axes = []
        labels = self._update_labels()
        for r in range(self.rows):
            if self._share_axes in {"x", "both"}:
                rel_height = 0.75 + 0.25 / self.rows / self._expand_top
                height = 0.75 / self.rows / rel_height / self._expand_top
                bottom_pad = 0.2 / self.rows / rel_height / self._expand_top
                bottom = bottom_pad + height * (self.rows - 1 - r)
            else:
                bottom_pad = 0.2 / self.rows
                height = 0.75 / self.rows / self._expand_top
                bottom = bottom_pad + (self.rows - 1 - r) / self.rows / self._expand_top
            for c in range(self.columns):
                if self._share_axes in {"y", "both"}:
                    rel_width = 0.75 + 0.25 / self.columns
                    width = 0.75 / self.columns / rel_width
                    left_pad = 0.2 / self.columns / rel_width
                    left = left_pad + width * c
                else:
                    left_pad = 0.2 / self.columns
                    width = 0.75 / self.columns
                    left = left_pad + c / self.columns
                axes.append(fig.add_axes((left, bottom, width, height)))
                if self._share_axes in {"x", "both"} and r != self.rows - 1:
                    axes[-1].set_xticklabels([])
                if self._share_axes in {"y", "both"} and c != 0:
                    axes[-1].set_yticklabels([])
                axes[-1].text(
                    self._pos[0],
                    self._pos[1],
                    labels[self.columns * r + c],
                    transform=axes[-1].transAxes,
                    **kwargs,
                )
        axes = self._maybe_columns_first(axes, transpose=False)
        return fig, axes


figure_grid = FigureGrid()



"""Create an airport-plot on a ``matplotlib.axes.Axes`` object."""

from collections.abc import Callable

import fppanalysis
import matplotlib as mpl
import numpy as np
import scipy
from numpy.typing import NDArray


class Airport:
    """Create an airport-plot."""

    def _setup(
        self,
        arr1: NDArray[np.float64],
        arr2: NDArray[np.float64],
        control: NDArray[np.float64],
        labels: tuple[str | None, str | None, str | None] = (
            "Arr2-Arr1",
            "Control",
            "Runway data",
        ),
        ax_labels: tuple[str, str, str] = (
            "X label",
            "Y label",
            "Distribution X label",
        ),
    ) -> None:
        self.arr1 = arr1
        self.arr2 = arr2
        self.control = control
        self.labels = labels
        self.ax_labels = ax_labels
        self.bins = 30
        self.dist_func_discrete = self.dist_func_discrete_pdf
        self.dist_func_continuous = self.dist_func_continuous_pdf

    def __call__(  # noqa: PLR0913
        self,
        arr1: NDArray[np.float64],
        arr2: NDArray[np.float64],
        control: NDArray[np.float64],
        ax: mpl.axes.Axes,
        labels: tuple[str | None, str | None, str | None] = (
            "Arr2-Arr1",
            "Control",
            "Runway data",
        ),
        ax_labels: tuple[str, str, str] = (
            "X label",
            "Y label",
            "Distribution X label",
        ),
    ) -> mpl.axes.Axes:
        """Create an airport-plot with default settings."""
        self.config()
        self._setup(arr1, arr2, control, labels, ax_labels)
        return self._plot(ax)

    def plot(  # noqa: PLR0913
        self,
        arr1: NDArray[np.float64],
        arr2: NDArray[np.float64],
        control: NDArray[np.float64],
        ax: mpl.axes.Axes,
        labels: tuple[str | None, str | None, str | None] = (
            "Arr2-Arr1",
            "Control",
            "Runway data",
        ),
        ax_labels: tuple[str, str, str] = (
            "X label",
            "Y label",
            "Distribution X label",
        ),
    ):
        """Create an airport-plot with the current settings."""
        self._setup(arr1, arr2, control, labels, ax_labels)
        self._plot(ax)

    def config(  # noqa: PLR0913
        self,
        *,
        bins: int = 30,
        distribution_scaling: float = 1.0,
        distribution_xline: float = 0.9,
        runway_end: float = 1.35,
        runway_start: float = -0.05,
        runway_data_kwargs: dict | None = None,
        dist_data_kwargs: dict | None = None,
        dist_ctrl_kwargs: dict | None = None,
    ) -> None:
        """Adjust the plot.

        Parameters
        ----------
        bins : int
            The number of bins to use when plotting the distributions.
        distribution_scaling : float
            Scale the distribution plot in its y-direction by this amount.
        distribution_xline : float
            Place the distribution plot's x-axis at this point along the runway.
        runway_end : float
            Place the runway end at this point relative to the data points.
        runway_start : float
            Place the runway start at this point relative to the data points.
        runway_data_kwargs : dict | None
            Keyword arguments to give to the runway scatter plot.
        dist_data_kwargs : dict | None
            Keyword arguments to give to the distribution plot's data.
        dist_ctrl_kwargs : dict | None
            Keyword arguments to give to the distribution plot's control.
        """
        if runway_data_kwargs is None:
            _runway_data_kwargs = {
                "marker": "*",
                "c": "magenta",
                "s": 12,
                "zorder": 10,
            }
        else:
            _runway_data_kwargs = runway_data_kwargs
        if dist_data_kwargs is None:
            _dist_data_kwargs = {
                "marker": "o",
                "c": "magenta",
                "s": 3,
                "zorder": 5,
            }
        else:
            _dist_data_kwargs = dist_data_kwargs
        if dist_ctrl_kwargs is None:
            _dist_ctrl_kwargs = {
                "marker": "d",
                "c": "orange",
                "s": 3,
                "zorder": 5,
            }
        else:
            _dist_ctrl_kwargs = dist_ctrl_kwargs
        self.distribution_scaling = distribution_scaling
        self.distribution_xline = distribution_xline
        self.runway_end = runway_end
        self.runway_start = runway_start
        self.runway_data_kwargs = _runway_data_kwargs
        self.dist_data_kwargs = _dist_data_kwargs
        self.dist_ctrl_kwargs = _dist_ctrl_kwargs

    @staticmethod
    def dist_func_discrete_pdf(
        n_bins: int, values: NDArray[np.float64]
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Probability distribution function estimate (discrete)."""
        out = fppanalysis.distribution(values, n_bins, ccdf=False)
        return out[2], out[0]

    @staticmethod
    def dist_func_continuous_pdf(
        x: NDArray[np.float64], y: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Probability distribution function estimate (continuous)."""
        fit_diff = scipy.stats.norm.fit(y)
        return scipy.stats.norm.pdf(x, *fit_diff)

    @staticmethod
    def dist_func_discrete_cdf(
        n_bins: int, values: NDArray[np.float64]
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Cumulative distribution function estimate (discrete)."""
        out = fppanalysis.distribution(values, n_bins, ccdf=False)
        return out[2], out[1]

    @staticmethod
    def dist_func_continuous_cdf(
        x: NDArray[np.float64], y: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        """Cumulative distribution function estimate (discrete)."""
        fit_diff = scipy.stats.norm.fit(y)
        return scipy.stats.norm.cdf(x, *fit_diff)

    def set_dist_func(
        self,
        dist_func_discrete: Callable[
            [int, NDArray[np.float64]],
            tuple[NDArray[np.float64], NDArray[np.float64]],
        ],
        dist_func_continuous: Callable[
            [NDArray[np.float64], NDArray[np.float64]],
            NDArray[np.float64],
        ],
    ) -> None:
        """Specify a custom distribution function to use in the plot.

        Two functions are expected, one discrete and one continuous. See the two
        implementaitons for PDF and CDF as examples.

        Parameters
        ----------
        dist_func_discrete : Callable[[int, NDArray[np.float64]], tuple[NDArray[np.float64], NDArray[np.float64]]]
            A function that takes in the number of bins and the values to use.
        dist_func_continuous : Callable[[NDArray[np.float64], NDArray[np.float64]], NDArray[np.float64]]
            A function that takes in the x-axis to to use and the y-values to fit to.
        """
        self.dist_func_discrete = dist_func_discrete  # type: ignore[assignment]
        self.dist_func_continuous = dist_func_continuous  # type: ignore[assignment]

    def _plot(
        self,
        ax: mpl.axes.Axes,
    ) -> mpl.axes.Axes:
        """Create an airport-plot on the given axes object."""
        ctrl_basis = -1 * self.control
        ctrl_bins, ctrl_dist = self.dist_func_discrete(self.bins, ctrl_basis)
        diff_basis = self.arr1 - self.arr2
        diff_bins, diff_dist = self.dist_func_discrete(self.bins, diff_basis)
        xy_range = np.sum([self.arr2, self.arr1], axis=0).flatten()
        # xmy_range = np.diff([self.arr1, self.arr2], axis=0).flatten()
        xy_range_diff = abs(xy_range.max() - xy_range.min())
        # xmy_range_diff = abs(xmy_range.max() - xmy_range.min())
        shift_x = (
            (
                (self.runway_end - self.runway_start) * self.distribution_xline
                + self.runway_start
            )
            * xy_range_diff
            + xy_range.min()
        ) / 2
        lims = (
            (self.runway_start * xy_range_diff + xy_range.min()) / 2,
            ((self.runway_end - 1) * xy_range_diff + xy_range.max()) / 2,
        )
        std = self.control.std() / 2
        ax.annotate(
            "",
            xy=(lims[1], lims[1]),
            xytext=(lims[0], lims[0]),
            arrowprops={
                "shrinkA": 0,
                "shrinkB": 0,
                "arrowstyle": "->",
                "lw": 0.7,
                "color": "grey",
            },
        )
        ax.fill(
            [lims[0] - std, lims[1] - std, lims[1] + std, lims[0] + std],
            [lims[0] + std, lims[1] + std, lims[1] - std, lims[0] - std],
            "grey",
            alpha=0.3,
        )
        ax.fill(
            [
                lims[0] - 2 * std,
                lims[1] - 2 * std,
                lims[1] + 2 * std,
                lims[0] + 2 * std,
            ],
            [
                lims[0] + 2 * std,
                lims[1] + 2 * std,
                lims[1] - 2 * std,
                lims[0] - 2 * std,
            ],
            "grey",
            alpha=0.3,
        )
        dist_ax_start = 1.05 * max(diff_bins.max(), ctrl_bins.max()) / 2
        dist_ax_end = 1.30 * min(ctrl_bins.min(), diff_bins.min()) / 2
        ax.annotate(
            self.ax_labels[2],
            xy=(dist_ax_start + shift_x, -dist_ax_start + shift_x),
            xytext=(dist_ax_end + shift_x, -dist_ax_end + shift_x),
            arrowprops={
                "shrinkA": 0,
                "shrinkB": 0,
                "lw": 0.7,
                "relpos": (1, 0),
                "color": "grey",
                "arrowstyle": "<-",
            },
            c="grey",
            ha="right",
            va="bottom",
            size=6,
        )
        bins_diff_scaled = diff_bins / 2
        bins_diff_xskew, bins_diff_yskew = (
            bins_diff_scaled + shift_x,
            bins_diff_scaled - shift_x,
        )
        bins_ctrl_scaled = ctrl_bins / 2
        bins_ctrl_xskew, bins_ctrl_yskew = (
            bins_ctrl_scaled + shift_x,
            bins_ctrl_scaled - shift_x,
        )
        ax.scatter(
            bins_diff_xskew + diff_dist * self.distribution_scaling,
            diff_dist * self.distribution_scaling - bins_diff_yskew,
            label=self.labels[0],
            **self.dist_data_kwargs,  # type: ignore[arg-type]
        )
        ax.scatter(
            bins_ctrl_xskew + ctrl_dist * self.distribution_scaling,
            ctrl_dist * self.distribution_scaling - bins_ctrl_yskew,
            label=self.labels[1],
            **self.dist_ctrl_kwargs,  # type: ignore[arg-type]
        )
        norm_diff = self.dist_func_continuous(diff_bins, diff_basis)
        norm_ctrl = self.dist_func_continuous(ctrl_bins, ctrl_basis)
        ax.plot(
            bins_diff_xskew + norm_ctrl * self.distribution_scaling,
            norm_ctrl * self.distribution_scaling - bins_diff_yskew,
            c=self.dist_data_kwargs["c"],
        )
        ax.plot(
            bins_ctrl_xskew + norm_diff * self.distribution_scaling,
            norm_diff * self.distribution_scaling - bins_ctrl_yskew,
            c=self.dist_ctrl_kwargs["c"],
        )
        ax.scatter(
            self.arr1,
            self.arr2,
            label=self.labels[2],
            **self.runway_data_kwargs,  # type: ignore[arg-type]
        )
        ax.set_xlabel(self.ax_labels[0])
        ax.set_ylabel(self.ax_labels[1])
        ax.legend(loc="lower right")
        ylim_min = ax.get_ylim()[0]
        ylim_max = -1.70 * min(ctrl_bins.min(), diff_bins.min()) / 2 + shift_x
        ylim_max = max(ylim_max, ax.get_ylim()[1])
        ax.set_ylim((ylim_min, ylim_max + 0.05 * (ylim_max - ylim_min)))
        return ax


airport = Airport()
